@@ -184,12 +184,15 @@ mapi_item_free (MapiItem *item, gpointer data)
 {
 	g_free (item->header.subject);
 	g_free (item->header.from);
-	/* g_free (item->header.from_email); */
-	/* g_free (item->header.from_type); */
+
 	g_free (item->header.to);
 	g_free (item->header.cc);
 	g_free (item->header.bcc);
 
+	g_free (item->header.references);
+	g_free (item->header.message_id);
+	g_free (item->header.in_reply_to);
+
 	exchange_mapi_util_free_attachment_list (&item->attachments);
 	exchange_mapi_util_free_stream_list (&item->generic_streams);
 }
@@ -259,6 +262,15 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 		case PR_MESSAGE_SIZE:
 			item->header.size = *(glong *)prop_data;
 			break;
+		case PR_INTERNET_MESSAGE_ID:
+			item->header.message_id = g_strdup (prop_data);
+			break;
+		case PR_INTERNET_REFERENCES:
+			item->header.references = g_strdup (prop_data);
+			break;
+		case PR_IN_REPLY_TO_ID:
+			item->header.in_reply_to = g_strdup (prop_data);
+			break;
 		case PR_MESSAGE_DELIVERY_TIME:
 			delivery_date = (struct FILETIME *) prop_data;
 			break;
@@ -321,6 +333,80 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	return TRUE;
 }
 
+static void
+mapi_set_message_id (CamelMapiMessageInfo *mapi_mi, const gchar *message_id)
+{
+	gchar *msgid;
+	guint8 *digest;
+	gsize length;
+	CamelMessageInfoBase *mi = &mapi_mi->info;
+
+	msgid = camel_header_msgid_decode (message_id);
+	if (msgid) {
+		GChecksum *checksum;
+
+		length = g_checksum_type_get_length (G_CHECKSUM_MD5);
+		digest = g_alloca (length);
+
+		checksum = g_checksum_new (G_CHECKSUM_MD5);
+		g_checksum_update (checksum, (guchar *) msgid, -1);
+		g_checksum_get_digest (checksum, digest, &length);
+		g_checksum_free (checksum);
+
+		memcpy(mi->message_id.id.hash, digest, sizeof(mi->message_id.id.hash));
+		g_free(msgid);
+	}
+
+}
+
+static void
+mapi_set_message_references (CamelMapiMessageInfo *mapi_mi, const gchar *references, const gchar *in_reply_to)
+{
+	struct _camel_header_references *refs, *irt, *scan;
+	guint8 *digest;
+	gint count;
+	gsize length;
+	CamelMessageInfoBase *mi = &mapi_mi->info;
+
+	refs = camel_header_references_decode (references);
+	irt = camel_header_references_inreplyto_decode (in_reply_to);
+	if (refs || irt) {
+		if (irt) {
+			/* The References field is populated from the "References" and/or "In-Reply-To"
+			   headers. If both headers exist, take the first thing in the In-Reply-To header
+			   that looks like a Message-ID, and append it to the References header. */
+
+			if (refs)
+				irt->next = refs;
+
+			refs = irt;
+		}
+
+		count = camel_header_references_list_size(&refs);
+		mi->references = g_malloc(sizeof(*mi->references) + ((count-1) * sizeof(mi->references->references[0])));
+
+		length = g_checksum_type_get_length (G_CHECKSUM_MD5);
+		digest = g_alloca (length);
+
+		count = 0;
+		scan = refs;
+		while (scan) {
+			GChecksum *checksum;
+
+			checksum = g_checksum_new (G_CHECKSUM_MD5);
+			g_checksum_update (checksum, (guchar *) scan->id, -1);
+			g_checksum_get_digest (checksum, digest, &length);
+			g_checksum_free (checksum);
+
+			memcpy(mi->references->references[count].id.hash, digest, sizeof(mi->message_id.id.hash));
+			count++;
+			scan = scan->next;
+		}
+		mi->references->size = count;
+		camel_header_references_list_clear(&refs);
+	}
+}
+
 static void
 mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **changeinfo,
 		   CamelException *ex, gboolean uid_flag) 
@@ -396,6 +482,12 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 			mi->info.date_sent = mi->info.date_received = item->header.recieved_time;
 			mi->info.size = (guint32) item->header.size;
 
+			/*Threading related properties*/
+			mapi_set_message_id (mi, item->header.message_id);
+			if (item->header.references || item->header.in_reply_to)
+				mapi_set_message_references (mi, item->header.references, item->header.in_reply_to);
+
+			/*Recipients*/
  			for (l = item->recipients; l; l=l->next) {
  				gchar *formatted_id = NULL;
 				const char *name, *display_name;
@@ -564,7 +656,7 @@ mapi_sync_deleted (CamelSession *session, CamelSessionThreadMsg *msg)
 		return;
 	}
 
-	camel_operation_start (NULL, _("Retriving message IDs from server for %s"), m->folder->name);
+	camel_operation_start (NULL, _("Retrieving message IDs from server for %s"), m->folder->name);
 
 	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 
@@ -814,6 +906,9 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		PR_SENT_REPRESENTING_EMAIL_ADDRESS,
 		PR_SENT_REPRESENTING_ADDRTYPE,
 		PR_LAST_MODIFICATION_TIME,
+		PR_INTERNET_MESSAGE_ID,
+		PR_INTERNET_REFERENCES,
+		PR_IN_REPLY_TO_ID,
 		PR_DISPLAY_TO,
 		PR_DISPLAY_CC,
 		PR_DISPLAY_BCC
@@ -957,6 +1052,11 @@ static const uint32_t camel_GetPropsList[] = {
 	PR_CONVERSATION_TOPIC, 
 	PR_CONVERSATION_TOPIC_UNICODE, 
 
+	/*Properties used for message threading.*/
+	PR_INTERNET_MESSAGE_ID,
+	PR_INTERNET_REFERENCES,
+	PR_IN_REPLY_TO_ID,
+
 	PR_BODY, 
 	PR_BODY_UNICODE, 
 	PR_HTML,
@@ -1142,6 +1242,15 @@ fetch_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 		case PR_MESSAGE_SIZE:
 			item->header.size = *(glong *)prop_data;
 			break;
+		case PR_INTERNET_MESSAGE_ID:
+			item->header.message_id = g_strdup (prop_data);
+			break;
+		case PR_INTERNET_REFERENCES:
+			item->header.references = g_strdup (prop_data);
+			break;
+		case PR_IN_REPLY_TO_ID:
+			item->header.in_reply_to = g_strdup (prop_data);
+			break;
 		case PR_MESSAGE_CLASS:
 		case PR_MESSAGE_CLASS_UNICODE:
 			msg_class = (const char *) prop_data;
@@ -1339,8 +1448,8 @@ mapi_populate_msg_body_from_item (CamelMultipart *multipart, MapiItem *item, Exc
 
 static CamelMimeMessage *
 mapi_folder_item_to_msg( CamelFolder *folder,
-		MapiItem *item,
-		CamelException *ex )
+			 MapiItem *item,
+			 CamelException *ex )
 {
 	CamelMimeMessage *msg = NULL;
 	CamelMultipart *multipart = NULL;
@@ -1350,12 +1459,9 @@ mapi_folder_item_to_msg( CamelFolder *folder,
 	/* char *body = NULL; */
 	ExchangeMAPIStream *body = NULL;
 	GSList *body_part_list = NULL;
-	const char *uid = NULL;
 
 	attach_list = item->attachments;
-
 	msg = camel_mime_message_new ();
-
 	multipart = camel_multipart_new ();
 
 	/*FIXME : Using set of default. Fix it during mimewriter*/
@@ -1366,7 +1472,7 @@ mapi_folder_item_to_msg( CamelFolder *folder,
 
 	camel_multipart_set_boundary (multipart, NULL);
 
-	camel_mime_message_set_message_id (msg, uid);
+	/* Handle Multipart */
 	body_part_list = item->msg.body_parts;
 	while (body_part_list){
 	       body = body_part_list->data;
@@ -1374,11 +1480,21 @@ mapi_folder_item_to_msg( CamelFolder *folder,
 	       body_part_list = g_slist_next (body_part_list);
 	}
 
+	/* Threading */
+	if (item->header.message_id)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "Message-Id", item->header.message_id);
+
+	if (item->header.references)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "References", item->header.references);
+
+	if (item->header.in_reply_to)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "In-Reply-To", item->header.in_reply_to);
 
 	/*Set recipient details*/
 	mapi_msg_set_recipient_list (msg, item);
 	mapi_populate_details_from_item (folder, msg, item);
 
+	/** Attachment Handling*/
 	if (attach_list) {
 		GSList *al = attach_list;
 		for (al = attach_list; al != NULL; al = al->next) {
@@ -1893,12 +2009,14 @@ camel_mapi_folder_new(CamelStore *store, const char *folder_name, const char *fo
 	}
 
 	/* set/load persistent state */
-	state_file = g_strdup_printf ("%s/cmeta", g_strdup_printf ("%s/%s",folder_dir, folder_name));
+	state_file = g_strdup_printf ("%s/%s/cmeta", folder_dir, folder_name);
 	camel_object_set(folder, NULL, CAMEL_OBJECT_STATE_FILE, state_file, NULL);
 	g_free(state_file);
 	camel_object_state_read(folder);
 
-	mapi_folder->cache = camel_data_cache_new (g_strdup_printf ("%s/%s",folder_dir, folder_name),0 ,ex);
+	state_file = g_strdup_printf ("%s/%s", folder_dir, folder_name);
+	mapi_folder->cache = camel_data_cache_new (state_file, 0, ex);
+	g_free (state_file);
 	if (!mapi_folder->cache) {
 		camel_object_unref (folder);
 		return NULL;
